@@ -1,10 +1,14 @@
 /* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 /*
- * Slabs memory allocation, based on powers-of-2
+ * Slabs memory allocation, based on powers-of-N. Slabs are up to 1MB in size
+ * and are divided into chunks. The chunk sizes start off at the size of the
+ * "item" structure plus space for a small key and value. They increase by
+ * a multiplier factor from there, up to half the maximum slab size. The last
+ * slab size is always 1MB, since that's the maximum item size allowed by the
+ * memcached protocol.
  *
- * $Id: slabs.c 393 2006-09-09 22:12:07Z bradfitz $
+ * $Id: slabs.c 369 2006-09-05 05:50:57Z bradfitz $
  */
-
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/time.h>
@@ -23,11 +27,12 @@
 
 #include "memcached.h"
 
-#define POWER_SMALLEST 3
-#define POWER_LARGEST  20
+#define POWER_SMALLEST 1
+#define POWER_LARGEST  200
 #define POWER_BLOCK 1048576
+#define CHUNK_ALIGN_BYTES (sizeof(void *))
 
-/* powers-of-2 allocation structures */
+/* powers-of-N allocation structures */
 
 typedef struct {
     unsigned int size;      /* sizes of items */
@@ -36,7 +41,7 @@ typedef struct {
     void **slots;           /* list of item ptrs */
     unsigned int sl_total;  /* size of previous array */
     unsigned int sl_curr;   /* first free slot */
- 
+
     void *end_page_ptr;         /* pointer to next free item at end of page, or 0 */
     unsigned int end_page_free; /* number of items remaining at end of last alloced page */
 
@@ -51,55 +56,73 @@ typedef struct {
 static slabclass_t slabclass[POWER_LARGEST+1];
 static size_t mem_limit = 0;
 static size_t mem_malloced = 0;
+static int power_largest;
 
+/*
+ * Figures out which slab class (chunk size) is required to store an item of
+ * a given size.
+ */
 unsigned int slabs_clsid(size_t size) {
-    int res = 1;
+    int res = POWER_SMALLEST;
 
     if(size==0)
         return 0;
-    size--;
-    while(size >>= 1)
-        res++;
-    if (res < POWER_SMALLEST) 
-        res = POWER_SMALLEST;
-    if (res > POWER_LARGEST)
-        res = 0;
+    while (size > slabclass[res].size)
+        if (res++ == power_largest)     /* won't fit in the biggest slab */
+            return 0;
     return res;
 }
 
-void slabs_init(size_t limit) {
-    int i;
-    int size=1;
+/*
+ * Determines the chunk sizes and initializes the slab class descriptors
+ * accordingly.
+ */
+void slabs_init(size_t limit, double factor) {
+    int i = POWER_SMALLEST - 1;
+    unsigned int size = sizeof(item) + settings.chunk_size;
+
+    /* Factor of 2.0 means use the default memcached behavior */
+    if (factor == 2.0 && size < 128)
+        size = 128;
 
     mem_limit = limit;
-    for(i=0; i<=POWER_LARGEST; i++, size*=2) {
+    memset(slabclass, 0, sizeof(slabclass));
+
+    while (++i < POWER_LARGEST && size <= POWER_BLOCK / 2) {
+        /* Make sure items are always n-byte aligned */
+        if (size % CHUNK_ALIGN_BYTES)
+            size += CHUNK_ALIGN_BYTES - (size % CHUNK_ALIGN_BYTES);
+
         slabclass[i].size = size;
-        slabclass[i].perslab = POWER_BLOCK / size;
-        slabclass[i].slots = 0;
-        slabclass[i].sl_curr = slabclass[i].sl_total = slabclass[i].slabs = 0;
-        slabclass[i].end_page_ptr = 0;
-        slabclass[i].end_page_free = 0;
-        slabclass[i].slab_list = 0;
-        slabclass[i].list_size = 0;
-        slabclass[i].killing = 0;
+        slabclass[i].perslab = POWER_BLOCK / slabclass[i].size;
+        size *= factor;
+        if (settings.verbose > 1) {
+            fprintf(stderr, "slab class %3d: chunk size %6d perslab %5d\n",
+                    i, slabclass[i].size, slabclass[i].perslab);
+        }
     }
 
+    power_largest = i;
+    slabclass[power_largest].size = POWER_BLOCK;
+    slabclass[power_largest].perslab = 1;
+
     /* for the test suite:  faking of how much we've already malloc'd */
     {
         char *t_initial_malloc = getenv("T_MEMD_INITIAL_MALLOC");
         if (t_initial_malloc) {
             mem_malloced = atol(getenv("T_MEMD_INITIAL_MALLOC"));
         }
+
     }
 
-    /* pre-allocate slabs by default, unless the environment variable
-       for testing is set to something non-zero */
+#ifndef DONT_PREALLOC_SLABS
     {
         char *pre_alloc = getenv("T_MEMD_SLABS_ALLOC");
         if (!pre_alloc || atoi(pre_alloc)) {
             slabs_preallocate(limit / POWER_BLOCK);
         }
     }
+#endif
 }
 
 void slabs_preallocate (unsigned int maxslabs) {
@@ -116,14 +139,14 @@ void slabs_preallocate (unsigned int maxslabs) {
         if (++prealloc > maxslabs)
             return;
         slabs_newslab(i);
-    } 
- 
+    }
+
 }
 
-static int grow_slab_list (unsigned int id) { 
+static int grow_slab_list (unsigned int id) {
     slabclass_t *p = &slabclass[id];
     if (p->slabs == p->list_size) {
-        unsigned int new_size =  p->list_size ? p->list_size * 2 : 16;
+        size_t new_size =  p->list_size ? p->list_size * 2 : 16;
         void *new_list = realloc(p->slab_list, new_size*sizeof(void*));
         if (new_list == 0) return 0;
         p->list_size = new_size;
@@ -134,21 +157,24 @@ static int grow_slab_list (unsigned int id) {
 
 int slabs_newslab(unsigned int id) {
     slabclass_t *p = &slabclass[id];
-    int num = p->perslab;
+#ifdef ALLOW_SLABS_REASSIGN
     int len = POWER_BLOCK;
+#else
+    int len = p->size * p->perslab;
+#endif
     char *ptr;
 
-    if (mem_limit && mem_malloced + len > mem_limit)
+    if (mem_limit && mem_malloced + len > mem_limit && p->slabs > 0)
         return 0;
 
     if (! grow_slab_list(id)) return 0;
-   
+
     ptr = malloc(len);
     if (ptr == 0) return 0;
 
     memset(ptr, 0, len);
     p->end_page_ptr = ptr;
-    p->end_page_free = num;
+    p->end_page_free = p->perslab;
 
     p->slab_list[p->slabs++] = ptr;
     mem_malloced += len;
@@ -159,7 +185,7 @@ void *slabs_alloc(size_t size) {
     slabclass_t *p;
 
     unsigned char id = slabs_clsid(size);
-    if (id < POWER_SMALLEST || id > POWER_LARGEST)
+    if (id < POWER_SMALLEST || id > power_largest)
         return 0;
 
     p = &slabclass[id];
@@ -171,7 +197,7 @@ void *slabs_alloc(size_t size) {
     mem_malloced += size;
     return malloc(size);
 #endif
-    
+
     /* fail unless we have space at the end of a recently allocated page,
        we have something on our freelist, or we could allocate a new page */
     if (! (p->end_page_ptr || p->sl_curr || slabs_newslab(id)))
@@ -200,8 +226,8 @@ void slabs_free(void *ptr, size_t size) {
     slabclass_t *p;
 
     assert(((item *)ptr)->slabs_clsid==0);
-    assert(id >= POWER_SMALLEST && id <= POWER_LARGEST);
-    if (id < POWER_SMALLEST || id > POWER_LARGEST)
+    assert(id >= POWER_SMALLEST && id <= power_largest);
+    if (id < POWER_SMALLEST || id > power_largest)
         return;
 
     p = &slabclass[id];
@@ -226,14 +252,14 @@ void slabs_free(void *ptr, size_t size) {
 
 char* slabs_stats(int *buflen) {
     int i, total;
-    char *buf = (char*) malloc(8192);
+    char *buf = (char*) malloc(power_largest * 200 + 100);
     char *bufcurr = buf;
 
     *buflen = 0;
     if (!buf) return 0;
 
     total = 0;
-    for(i = POWER_SMALLEST; i <= POWER_LARGEST; i++) {
+    for(i = POWER_SMALLEST; i <= power_largest; i++) {
         slabclass_t *p = &slabclass[i];
         if (p->slabs) {
             unsigned int perslab, slabs;
@@ -257,7 +283,13 @@ char* slabs_stats(int *buflen) {
     return buf;
 }
 
-/* 1 = success
+#ifdef ALLOW_SLABS_REASSIGN
+/* Blows away all the items in a slab class and moves its slabs to another
+   class. This is only used by the "slabs reassign" command, for manual tweaking
+   of memory allocation. It's disabled by default since it requires that all
+   slabs be the same size (which can waste space for chunk size mantissas of
+   other than 2.0).
+   1 = success
    0 = fail
    -1 = tried. busy. send again shortly. */
 int slabs_reassign(unsigned char srcid, unsigned char dstid) {
@@ -266,10 +298,10 @@ int slabs_reassign(unsigned char srcid, unsigned char dstid) {
     void *iter;
     int was_busy = 0;
 
-    if (srcid < POWER_SMALLEST || srcid > POWER_LARGEST ||
-        dstid < POWER_SMALLEST || dstid > POWER_LARGEST)
+    if (srcid < POWER_SMALLEST || srcid > power_largest ||
+        dstid < POWER_SMALLEST || dstid > power_largest)
         return 0;
-   
+
     p = &slabclass[srcid];
     dp = &slabclass[dstid];
 
@@ -280,12 +312,12 @@ int slabs_reassign(unsigned char srcid, unsigned char dstid) {
     /* fail if dst is still growing or we can't make room to hold its new one */
     if (dp->end_page_ptr || ! grow_slab_list(dstid))
         return 0;
-        
+
     if (p->killing == 0) p->killing = 1;
 
     slab = p->slab_list[p->killing-1];
     slab_end = slab + POWER_BLOCK;
-    
+
     for (iter=slab; iter<slab_end; iter+=p->size) {
         item *it = (item *) iter;
         if (it->slabs_clsid) {
@@ -293,7 +325,7 @@ int slabs_reassign(unsigned char srcid, unsigned char dstid) {
             item_unlink(it);
         }
     }
-    
+
     /* go through free list and discard items that are no longer part of this slab */
     {
         int fi;
@@ -314,12 +346,11 @@ int slabs_reassign(unsigned char srcid, unsigned char dstid) {
     dp->slab_list[dp->slabs++] = slab;
     dp->end_page_ptr = slab;
     dp->end_page_free = dp->perslab;
-
     /* this isn't too critical, but other parts of the code do asserts to
        make sure this field is always 0.  */
     for (iter=slab; iter<slab_end; iter+=dp->size) {
         ((item *)iter)->slabs_clsid = 0;
     }
-
     return 1;
 }
+#endif
